#![allow(dead_code)]
use serde::{Deserialize, Serialize};
use std::sync::OnceLock;

use utoipa::{OpenApi, ToSchema};

#[derive(OpenApi)]
#[openapi(components(schemas(
    RegularSchema,
    NestedSchema,
    InnerSchema,
    InnerInnerSchema,
    Options,
    SerdeSchema,
    EnumType,
    TupleSchema,
    TupleStruct,
)))]
struct ApiDocs;

#[derive(ToSchema)]
struct RegularSchema {
    #[schema(min_length = 1, max_length = 5)]
    field: String,
    #[schema(minimum = 5, maximum = 10)]
    field2: i64,
    field3: bool,
    field4: Vec<String>,
}

#[derive(ToSchema)]
struct NestedSchema {
    #[schema(min_length = 1, max_length = 5)]
    field: String,
    nested: InnerSchema,
}

#[derive(ToSchema)]
struct InnerSchema {
    field: bool,
    #[schema(minimum = 5, maximum = 10)]
    field2: i32,
    nested: InnerInnerSchema,
}

#[derive(ToSchema)]
struct InnerInnerSchema {
    #[schema(pattern = "^test$")]
    field: String,
    field2: Options,
}

#[derive(ToSchema)]
enum Options {
    One,
    Two,
    Three,
}

#[derive(ToSchema)]
struct TupleSchema {
    inner: TupleStruct,
}

#[derive(ToSchema)]
struct TupleStruct(Options);

#[derive(Debug, Serialize, Deserialize, ToSchema)]
pub struct SerdeSchema {
    #[serde(flatten)]
    flat: EnumType,
    #[schema(min_length = 1)]
    regular: String,
}

#[derive(Debug, Serialize, Deserialize, ToSchema)]
#[serde(tag = "tag")]
#[serde(rename_all = "lowercase")]
pub enum EnumType {
    First { field: String },
    Second { field: String },
    None,
}

fn get_schemas() -> &'static serde_json::Value {
    static SCHEMAS: OnceLock<serde_json::Value> = OnceLock::new();
    SCHEMAS.get_or_init(|| {
        serde_json::to_value(
            ApiDocs::openapi()
                .components
                .expect("no components found")
                .schemas,
        )
        .unwrap()
    })
}

#[test]
fn init_validator_regular() {
    let schemas = get_schemas();
    schema_police::SchemaInspector::new_infer::<RegularSchema>(schemas, false).unwrap();
}

#[test]
fn init_validator_nested() {
    let schemas = get_schemas();
    schema_police::SchemaInspector::new_infer::<NestedSchema>(schemas, true).unwrap();
}

#[test]
fn init_validator_serde() {
    let schemas = get_schemas();
    schema_police::SchemaInspector::new_infer::<SerdeSchema>(schemas, true).unwrap();
}

#[test]
fn validate_regular_schema_ok() {
    let schemas = get_schemas();
    let validator =
        schema_police::SchemaInspector::new_infer::<RegularSchema>(schemas, false).unwrap();
    let result = validator.inspect(&serde_json::json!({
        "field": "field",
        "field2": 8,
        "field3": true,
        "field4": [""],
    }));
    assert_eq!(result, Ok(()));
}

#[test]
fn validate_regular_schema_err() {
    let schemas = get_schemas();
    let validator =
        schema_police::SchemaInspector::new_infer::<RegularSchema>(schemas, false).unwrap();
    let result = validator.inspect(&serde_json::json!({
        "field": "eld",
        "field2": 0,
        "field3": true,
        "field4": [""]
    }));
    assert!(result.is_err());
}

#[test]
fn validate_nested_schema_ok() {
    let schemas = get_schemas();
    let validator =
        schema_police::SchemaInspector::new_infer::<NestedSchema>(schemas, true).unwrap();
    let result = validator.inspect(&serde_json::json!({
        "field": "field",
        "nested": {
            "field": true,
            "field2": 6,
            "nested": {
                "field": "test",
                "field2": "One"
            }
        }
    }));
    assert_eq!(result, Ok(()));
}

#[test]
fn validate_nested_schema_err() {
    let schemas = get_schemas();
    let validator =
        schema_police::SchemaInspector::new_infer::<NestedSchema>(schemas, true).unwrap();
    let result = validator.inspect(&serde_json::json!({
        "field": "field",
        "nested": {
            "field": true,
            "field2": 6,
            "nested": {
                "field": "",
                "field2": "test"
            }
        }
    }));
    assert!(result.is_err())
}

#[test]
fn validate_serde_schema_ok() {
    let schemas = get_schemas();
    let validator =
        schema_police::SchemaInspector::new_infer::<SerdeSchema>(schemas, true).unwrap();
    let result = validator.inspect(&serde_json::json!({
        "regular": "field",
        "tag": "first",
        "field": ""
    }));
    assert_eq!(result, Ok(()));
}

#[test]
fn validate_serde_schema_err() {
    let schemas = get_schemas();
    let validator =
        schema_police::SchemaInspector::new_infer::<SerdeSchema>(schemas, true).unwrap();
    let result = validator.inspect(&serde_json::json!({
        "regular": "field",
        "tag": "huh",
        "field": ""
    }));
    assert!(result.is_err());
}

#[test]
fn validate_tuple_schema() {
    let schemas = get_schemas();
    let validator =
        schema_police::SchemaInspector::new_infer::<TupleSchema>(schemas, true).unwrap();
    let result = validator.inspect(&serde_json::json!({
        "inner": "One",
    }));
    assert_eq!(result, Ok(()));
    let result = validator.inspect(&serde_json::json!({
        "inner": "Two",
    }));
    assert_eq!(result, Ok(()));
    let result = validator.inspect(&serde_json::json!({
        "inner": "Three",
    }));
    assert_eq!(result, Ok(()));
}
